/* Copyright (c) 2004 Alessandro Rubini. GNU GPL 2 or later */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kobject.h>
#include <linux/sysfs.h>

MODULE_LICENSE("GPL");

/* the toplevel kint structure */
struct kint {
	struct kobject kobj;
	int numbers[3]; /* min, val, max */
};

/* macros to extract the container structure from the generic one */
#define to_kint(obj)      container_of(obj, struct kint, kobj)
#define to_int_attr(attr) container_of(attr, struct kint_attribute, attr)

/* Functions to read and write one signed integer in a range */
ssize_t kint_one_show(int *what, char *buf)
{
	return sprintf(buf, "%i\n", *what);
}
ssize_t kint_one_store(int *what, const char *buf, size_t count)
{
	int result;
	char s[4];

	if (sscanf(buf, "%i %2s", &result, s) != 1)
		return -EINVAL;
	if (result < what[-1] || result > what[1])
		return -EINVAL;
	*what = result;
	return strlen(buf);
}

/* This customized attribute allows to read min/max and read/write val */
struct kint_attribute {
	struct attribute attr;
	int index;
	ssize_t (*show)(int *, char *);
	ssize_t (*store)(int *, const char *, size_t count);
};

struct kint_attribute kint_attr_min = {
        .attr    = {.name = "min", .mode = S_IRUGO },
	.index   = 0,
        .show    = kint_one_show
};
struct kint_attribute kint_attr_val = {
        .attr    = {.name = "val", .mode = S_IRUGO | S_IWUGO },
	.index   = 1,
        .show    = kint_one_show,
	.store   = kint_one_store
};
struct kint_attribute kint_attr_max = {
        .attr    = {.name = "max", .mode = S_IRUGO },
	.index   = 2,
        .show    = kint_one_show
};

struct attribute * kint_default_attrs[] = {
        &kint_attr_min.attr,
        &kint_attr_val.attr,
        &kint_attr_max.attr,
        NULL
};


/* top level show/store functions, called by sysfs */
ssize_t kint_show(struct kobject * kobj, struct attribute * attr, char * buf)
{
	struct kint *ki = to_kint(kobj);
	struct kint_attribute *ka = to_int_attr(attr);
	ssize_t ret = 0;

	if (ka->show)
		ret = ka->show(&ki->numbers[ka->index], buf);
	return ret;
}
ssize_t kint_store(struct kobject * kobj, struct attribute * attr, 
               const char * buf, size_t count)
{
	struct kint *ki = to_kint(kobj);
	struct kint_attribute *ka = to_int_attr(attr);
	ssize_t ret = 0;

	if (ka->store)
		ret = ka->store(&ki->numbers[ka->index], buf, count);
	return ret;
}

struct sysfs_ops kint_sysfs_ops = {
	.show  = kint_show,
	.store = kint_store,
};

/* our object, dynamically allocated */
struct kint *kint;

/* our release method */
void kint_release(struct kobject *kobj)
{
	struct kint *ki = to_kint(kobj);
	kfree(ki);
	kint = NULL; /* no doubt it's this one */
}

/* the ktype is referenced by kobject */
struct kobj_type ktype_int = {
	.release       = kint_release,
	.sysfs_ops     = &kint_sysfs_ops,
	.default_attrs = kint_default_attrs,
};

/* module initialization and cleanup */
int int_init(void)
{
	int i, ret;

	kint = kmalloc(sizeof(*kint), GFP_KERNEL);
	if (!kint) return -ENOMEM;
	memset(kint, 0, sizeof(*kint));
	kobject_init(&kint->kobj);
	kint->numbers[0] = 0;
	kint->numbers[1] = 1;
	kint->numbers[2] = 5;
	kint->kobj.ktype = &ktype_int;

	ret = kobject_set_name(&kint->kobj, "sample-int-range");
	if (!ret) ret = kobject_add(&kint->kobj);
	if (!ret) {
		for (i=0; i<3; i++)
			sysfs_create_file(&kint->kobj,
					  kint_default_attrs[i]);
	}

	if (ret)
		kfree(kint);
	return ret;
}

void int_exit(void)
{
	int i;

	for (i=0; i<3; i++)
		sysfs_remove_file(&kint->kobj,
				  kint_default_attrs[i]);
	kobject_del(&kint->kobj);
	kobject_put(&kint->kobj);
	/* wait untile the kobj is freed */
	
	while (kint) {
		schedule_timeout(HZ/2);
	}
	return;
}

module_init(int_init);
module_exit(int_exit);
